@@ -3,6 +3,7 @@
 from copy import deepcopy
 from typing import TYPE_CHECKING, Generic, TypeVar
 from weakref import WeakValueDictionary, ref
+from threading import RLock
 
 from six import with_metaclass
 
@@ -12,19 +13,19 @@ from .generic_meta import GenericMeta
 if TYPE_CHECKING:
     from typing import Dict, Any, Optional, Type
 
-__all__ = ["WeakReference"]
+__all__ = ["WeakReference", "UniqueHashWeakReference"]
 
 
 T = TypeVar("T")  # Any type.
 
-_DEAD_REF = ref(
-    type("Dead", (object,), {"__slots__": ("__weakref__",)})()
-)
+_DEAD_REF = ref(type("Dead", (object,), {"__slots__": ("__weakref__",)})())
 
 _cache = (
     {}
 )  # type: Dict[Type["WeakReference"], WeakValueDictionary[int, "WeakReference"]]
 
+_lock = RLock()
+
 
 def _reduce_dead(cls):
     self = super(WeakReference, cls).__new__(cls)  # type: ignore
@@ -58,26 +59,27 @@ class WeakReference(with_metaclass(GenericMeta, Generic[T])):
     @staticmethod
     def __new__(cls, obj=None):
         # type: (Type[WeakReference[T]], T) -> WeakReference[T]
-        cache = _cache.setdefault(cls, WeakValueDictionary())
-        if obj is None:
-            obj_ref = _DEAD_REF
-        else:
-            obj_ref = ref(obj)
+        with _lock:
+            cache = _cache.setdefault(cls, WeakValueDictionary())
+            if obj is None:
+                obj_ref = _DEAD_REF
+            else:
+                obj_ref = ref(obj)
 
-        obj_ref_id = id(obj_ref)
-        try:
-            self = cache[obj_ref_id]
-        except KeyError:
-            pass
-        else:
+            obj_ref_id = id(obj_ref)
             try:
-                if self.__ref is obj_ref:
-                    return self
-            except ReferenceError:
+                self = cache[obj_ref_id]
+            except KeyError:
                 pass
+            else:
+                try:
+                    if self.__ref is obj_ref:
+                        return self
+                except ReferenceError:
+                    pass
 
-        self = cache[obj_ref_id] = super(WeakReference, cls).__new__(cls)
-        return self
+            self = cache[obj_ref_id] = super(WeakReference, cls).__new__(cls)
+            return self
 
     def __init__(self, obj=None):
         # type: (T) -> None
@@ -98,7 +100,7 @@ class WeakReference(with_metaclass(GenericMeta, Generic[T])):
         except KeyError:
             cls = type(self)
             obj = self()
-            if self is WeakReference():
+            if self is type(self)():
                 assert obj is None
                 copy = memo[self_id] = self
             elif obj is None:
@@ -187,9 +189,32 @@ class WeakReference(with_metaclass(GenericMeta, Generic[T])):
     def __reduce__(self):
         cls = type(self)
         obj = self()
-        if self is WeakReference():
+        if self is type(self)():
             assert obj is None
             return cls, (None,)
         if obj is None:
             return _reduce_dead, (cls,)
         return cls, (obj,)
+
+
+class UniqueHashWeakReference(WeakReference, Generic[T]):
+    """
+    Weak Reference-like object that supports pickling and has a unique id-based hash.
+
+    .. code:: python
+
+        >>> import pickle
+        >>> from basicco.utils.weak_reference import UniqueHashWeakReference
+        >>> class MyClass(object):
+        ...     __hash__ = None
+        ...
+        >>> strong = MyClass()
+        >>> weak = UniqueHashWeakReference(strong)
+        >>> hash(weak) == object.__hash__(weak)
+        True
+    """
+
+    __slots__ = ()
+
+    def __hash__(self):
+        return object.__hash__(self)
